;; Generated package description from el-search.el
(define-package "el-search" "1.10.2" "Expression based interactive search for Emacs Lisp" '((emacs "25") (stream "2.2.4") (cl-print "1.0")) :url "http://elpa.gnu.org/packages/el-search.html" :keywords '("lisp"))
